//
// mpiJava version : Sung-Hoon Ko
//                   June 1998
// Northeast Parallel Architectures Center at Syracuse University
//

import java.io.*;

import mpi.*;

class PingPong {

  static final int  PINGPONGS = 24;
  static final int  SAMPLES   = 100;
  static final int  LOG2N_MAX = 14;
  static final int  N_MAX     = ( 1 << (LOG2N_MAX-3) );

  public static void main(String[] args) {
     
    double      t_call;
    int         i,j,k, j_pe, log2nbyte, nbyte, ns, n_pp, my_pe, npes;
    int         n_init, n_mess;
    //    String      Title ="Single Messages --- MPI_JAVA on Solaris(SM) ";
    String      Title ="Single Messages --- MPI_JAVA";
    PrintStream pfout = null;
    Status      status;

    MPI.Init(args);

    my_pe = MPI.COMM_WORLD.Rank();
    npes  = MPI.COMM_WORLD.Size();

    double  tf[] = new double [SAMPLES + 1];

    System.out.println("\n"+MPI.Get_processor_name());


//***************************************************************************
//              FLOAT with [1][nbyte*nbyte]                                  
//***************************************************************************
    //    LOG2N_MAX = 12;
    System.out.println("********************************************\n");
    System.out.println("Start Sending (float) A[1][nbyte*nbyte] : MPI_OBJECT ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }


    k = 2 ;
    for (log2nbyte = 14; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << 10 );
	//nbyte = ( 1 << log2nbyte );
	
	//if( nbyte > 128 ) { 
	//  nbyte = k * 128;
	//  k++;	    
	//}    

	float A[][] = new float [1][nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[0][i] = i;
      }
          
      for (j_pe = 2; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, 1, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,1, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with (float)[1][nbyte*nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();

    //******************************************************
//***************************************************************************
//                 FLOAT with [nbyte][nbyte]                                
//***************************************************************************
    //******************************************************
    //    LOG2N_MAX = 12;
    System.out.println("********************************************");
    System.out.println("Start Sending (float) A[nbyte][nbyte] : MPI_OBJECT ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }



    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    


      float A[][] = new float [nbyte][nbyte];
      
      for (i = 0; i < nbyte; i++) {
         for(j = 0; j < nbyte; j++)
	    A[i][j] = i;
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, nbyte, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with (float)[nbyte][nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();




    //******************************************************
    //     Serialize Time FLOAT with  [1][nbyte]  
    //******************************************************
    System.out.println("##############################################");;
    System.out.println("********************************************\n");
    System.out.println("Serialize Time ...");
    System.out.println("Start (float) A[1][nbyte*nbyte] ...");


    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = PINGPONGS;
    n_mess = PINGPONGS;
    if (my_pe == 0) {
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    

      float A[][] = new float [1][nbyte*nbyte];

      for (i = 0; i < nbyte*nbyte; i++) {
	    A[0][i] = i;
      }
      //for (j_pe = 0; j_pe <= npes; ++j_pe) {
	  MPI.COMM_WORLD.Barrier();
	  tf[0] = MPI.Wtime();
	  int len = 0;
	  for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
                  byte[] byte_buf = MPI.COMM_WORLD.Object_Serialize(A,0,1,MPI.OBJECT);
		  len = byte_buf.length;    
              }
	      tf[ns+1] = MPI.Wtime();
	  }
	  if(my_pe==0)
	      Table.table_body_se_dese(my_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout,len);
	  //}
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":FLOAT with [1][nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();

    //******************************************************
//***************************************************************************
//                   FLOAT with [1][nbyte]                                   
//***************************************************************************
    //******************************************************
    System.out.println("********************************************\n");
    System.out.println("Deserialize Time ...");
    System.out.println("Start (float) with [1][nbyte*nbyte] ...");


    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = PINGPONGS;
    n_mess = PINGPONGS;    
    if (my_pe ==0)
       Table.table_top(npes,n_init, n_mess, SAMPLES, t_call, Title);


    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    

      float A[][] = new float [1][nbyte*nbyte];

      for (i = 0; i < nbyte*nbyte; i++) {
	    A[0][i] = i;
      }
      byte[] byte_buf = MPI.COMM_WORLD.Object_Serialize(A,0,1,MPI.OBJECT);

      //  for (j_pe = 0; j_pe <= npes; ++j_pe) {
	  MPI.COMM_WORLD.Barrier();
	  tf[0] = MPI.Wtime();
	  int len = 0;
	  for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
                  MPI.COMM_WORLD.Object_Deserialize(A,byte_buf,0,1,MPI.OBJECT);
		  len = byte_buf.length;    
              }
	      tf[ns+1] = MPI.Wtime();
	  }
	  if(my_pe==0)
	      Table.table_body_se_dese(my_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout,len);
	  //}
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":FLOAT with [1][nbyte*nbyte] -- Finished");

    MPI.COMM_WORLD.Barrier();



    //******************************************************
//***************************************************************************
//                FLOAT [nbyte*nbyte]                                      
//***************************************************************************
    //******************************************************
    System.out.println("Start Sending (float) A[N*N]: MPI_FLOAT ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }
    
    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    
	//for (log2nbyte = 1; log2nbyte <= 8; ++log2nbyte) {
	//nbyte = log2nbyte * 128;

	float A[] = new float [nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[i] = i ;
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	  MPI.COMM_WORLD.Barrier();
	  tf[0] = MPI.Wtime();
	  for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
		  if (my_pe == j_pe) {
		      MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.FLOAT, 0, 10);
		      status = MPI.COMM_WORLD.Recv(A, 0, nbyte*nbyte, MPI.FLOAT, 0, 20);
		  }
		  if (my_pe == 0) {
		      status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte*nbyte, MPI.FLOAT, j_pe, 10);
		      MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.FLOAT, j_pe, 20);
		  }
	      }
	      tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	      Table.table_body(j_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
	  
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":FLOAT -- Finished");
    MPI.COMM_WORLD.Barrier();



    //******************************************************
//***************************************************************************
//                   BYTE [nbyte*nbyte]                                      
//***************************************************************************
    //******************************************************
    //    LOG2N_MAX = 14;
    System.out.println("Start Sending (byte) A[N*N]: MPI_BYTE ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }
    
    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    
	//for (log2nbyte = 1; log2nbyte <= 8; ++log2nbyte) {
	//nbyte = log2nbyte * 128;

	byte A[] = new byte [nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[i] = (byte)'0'; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	  MPI.COMM_WORLD.Barrier();
	  tf[0] = MPI.Wtime();
	  for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
		  if (my_pe == j_pe) {
		      MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.BYTE, 0, 10);
		      status = MPI.COMM_WORLD.Recv(A, 0, nbyte*nbyte, MPI.BYTE, 0, 20);
		  }
		  if (my_pe == 0) {
		      status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte*nbyte, MPI.BYTE, j_pe, 10);
		      MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.BYTE, j_pe, 20);
		  }
	      }
	      tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	      Table.table_body(j_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE -- Finished");
    MPI.COMM_WORLD.Barrier();

    //******************************************************
//***************************************************************************
//                   BYTE with [1][nbyte]                                    
//***************************************************************************
    //******************************************************
    System.out.println("********************************************\n");
    System.out.println("Start Sending (byte) A[1][nbyte*nbyte] :MPI_OBJECT ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }


    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    

	byte A[][] = new byte [1][nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[0][i] = (byte)'0'; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, 1, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,1, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with [1][nbyte*nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();

    //******************************************************
//***************************************************************************
//                   BYTE with [nbyte][nbyte]                                
//***************************************************************************
    //******************************************************
    System.out.println("********************************************");
    System.out.println("Start Sending (byte) A[nbyte][nbyte] : MPI_OBJECT ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }



    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    


      byte A[][] = new byte [nbyte][nbyte];
      
      for (i = 0; i < nbyte; i++) {
         for(j = 0; j < nbyte; j++)
	    A[i][j] = (byte)'0'; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, nbyte, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with [nbyte][nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();

    //******************************************************
    //******************************************************
    //******************************************************

    //******************************************************************
//                   BYTE with  [1][nbyte]                               
    //******************************************************************
    //******************************************************
    System.out.println("##############################################");;
    System.out.println("********************************************\n");
    System.out.println("Serialize Time ...");
    System.out.println("Start (byte) A[1][nbyte*nbyte] ...");


    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = PINGPONGS;
    n_mess = PINGPONGS;
    if (my_pe == 0) {
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    

      byte A[][] = new byte [1][nbyte*nbyte];

      for (i = 0; i < nbyte*nbyte; i++) {
	    A[0][i] = (byte)'0'; //(double) 1. / (i + 1);
      }
      //for (j_pe = 0; j_pe <= npes; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
            int len = 0;
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
                  byte[] byte_buf = MPI.COMM_WORLD.Object_Serialize(A,0,1,MPI.OBJECT);
              len = byte_buf.length;    
              }
	      tf[ns+1] = MPI.Wtime();
	    }
	    if(my_pe==0)
        Table.table_body_se_dese(my_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout,len);
	    //}
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with [1][nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();

    //******************************************************
//***************************************************************************
//                   BYTE with [1][nbyte]                                    
//***************************************************************************
    //******************************************************
    System.out.println("********************************************\n");
    System.out.println("Deserialize Time ...");
    System.out.println("Start BYTE with [1][nbyte*nbyte] ...");


    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = PINGPONGS;
    n_mess = PINGPONGS;    
    if (my_pe ==0)
       Table.table_top(npes,n_init, n_mess, SAMPLES, t_call, Title);


    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    

      byte A[][] = new byte [1][nbyte*nbyte];

      for (i = 0; i < nbyte*nbyte; i++) {
	    A[0][i] = (byte)'0'; //(double) 1. / (i + 1);
      }
      byte[] byte_buf = MPI.COMM_WORLD.Object_Serialize(A,0,1,MPI.OBJECT);

      //      for (j_pe = 0; j_pe <= npes; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
            int len = 0;
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
                  MPI.COMM_WORLD.Object_Deserialize(A,byte_buf,0,1,MPI.OBJECT);
              len = byte_buf.length;    
              }
	      tf[ns+1] = MPI.Wtime();
	      }
	      if(my_pe==0)
        Table.table_body_se_dese(my_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout,len);
	      //}
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with [1][nbyte*nbyte] -- Finished");



    //******************************************************
//***************************************************************************
//                   BYTE with  [nbyte][nbyte]                               
//***************************************************************************
    //******************************************************
    System.out.println("##############################################");;
    System.out.println("********************************************\n");
    System.out.println("Serialize Time ...");
    System.out.println("Start (byte) A[nbyte][nbyte] ...");


    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = PINGPONGS;
    n_mess = PINGPONGS;
    if (my_pe == 0) {
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    

      byte A[][] = new byte [nbyte][nbyte];

      for (i = 0; i < nbyte; i++) {
	  for (j = 0; j < nbyte; i++) {
	    A[j][i] = (byte)'0'; //(double) 1. / (i + 1);
	  }
      }
      //      for (j_pe = 1; j_pe <= npes; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
            int len = 0;
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
                  byte[] byte_buf = MPI.COMM_WORLD.Object_Serialize(A,0,nbyte,MPI.OBJECT);
              len = byte_buf.length;    
              }
	      tf[ns+1] = MPI.Wtime();
	    }
	    if(my_pe==0)
        Table.table_body_se_dese(my_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout,len);
	    //}
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with [nbyte][nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();

    //********************************************************
//***************************************************************************
//                   BYTE with [nbyte][nbyte]                                    
//***************************************************************************
    //******************************************************
    System.out.println("********************************************\n");
    System.out.println("Deserialize Time ...");
    System.out.println("Start BYTE with [nbyte][nbyte] ...");


    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = PINGPONGS;
    n_mess = PINGPONGS;    
    if (my_pe ==0)
       Table.table_top(npes,n_init, n_mess, SAMPLES, t_call, Title);


    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    

      byte A[][] = new byte [nbyte][nbyte];


      for (i = 0; i < nbyte; i++) {
	  for (j = 0; j < nbyte; i++) {
	    A[j][i] = (byte)'0'; //(double) 1. / (i + 1);
	  }
      }

      byte[] byte_buf = MPI.COMM_WORLD.Object_Serialize(A,0,nbyte,MPI.OBJECT);

      //      for (j_pe = 0; j_pe <= npes; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
            int len = 0;
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
                  MPI.COMM_WORLD.Object_Deserialize(A,byte_buf,0,nbyte,MPI.OBJECT);
              len = byte_buf.length;    
              }
	      tf[ns+1] = MPI.Wtime();
	      }
	      if(my_pe==0)
        Table.table_body_se_dese(my_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout,len);
	      //}
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with [nbyte][nbyte] -- Finished");

    MPI.COMM_WORLD.Barrier();
    //******************************************************
//***************************************************************************
//                   INTE with  [nbyte][nbyte]                               
////*************************************************************************
    //******************************************************
    System.out.println("##############################################");;
    System.out.println("********************************************\n");
    System.out.println("Serialize Time ...");
    System.out.println("Start (int) A[nbyte]nbyte] ...");


    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = PINGPONGS;
    n_mess = PINGPONGS;
    if (my_pe == 0) {
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    

      int A[][] = new int [nbyte][nbyte];

      for (i = 0; i < nbyte; i++) {
	  for (j = 0; j < nbyte; i++) {
	    A[j][i] = i;
	  }
      }
      //      for (j_pe = 0; j_pe <= npes; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
            int len = 0;
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
                  byte[] byte_buf = MPI.COMM_WORLD.Object_Serialize(A,0,nbyte,MPI.OBJECT);
              len = byte_buf.length;    
              }
	      tf[ns+1] = MPI.Wtime();
	    }
	    if(my_pe==0)
        Table.table_body_se_dese(my_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout,len);
	    //}
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":INT with [nbyte][nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();

    //******************************************************
////*************************************************************************
//                   INT with [nbyte][nbyte]                                    
////*************************************************************************
    //******************************************************
    System.out.println("********************************************\n");
    System.out.println("Deserialize Time ...");
    System.out.println("Start INT with [nbyte][nbyte] ...");


    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = PINGPONGS;
    n_mess = PINGPONGS;    
    if (my_pe ==0)
       Table.table_top(npes,n_init, n_mess, SAMPLES, t_call, Title);


    k = 2 ;
    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
	nbyte = ( 1 << log2nbyte );
	
	if( nbyte > 128 ) { 
	    nbyte = k * 128;
	    k++;
	}    

      int A[][] = new int [nbyte][nbyte];


      for (i = 0; i < nbyte; i++) {
	  for (j = 0; j < nbyte; i++) {
	    A[j][i] = i;
	  }
      }

      byte[] byte_buf = MPI.COMM_WORLD.Object_Serialize(A,0,nbyte,MPI.OBJECT);

      //      for (j_pe = 0; j_pe <= npes; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
            int len = 0;
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
                  MPI.COMM_WORLD.Object_Deserialize(A,byte_buf,0,nbyte,MPI.OBJECT);
              len = byte_buf.length;    
              }
	      tf[ns+1] = MPI.Wtime();
	      }
	      if(my_pe==0)
        Table.table_body_se_dese(my_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout,len);
	      //}
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":INT with [nbyte][nbyte] -- Finished");

  
    MPI.Finalize();    
  }

}
